// This may look like C code, but it is really -*- C++ -*-

/* $Id: ASgFuncs3.cc,v 1.2 1994/06/07 21:34:23 Andrew-Mauer Exp $ */

#include <assert.h>
#include "sage++user.h"
#include "iterator-expr.h"
#include "ASgFuncs.h"
#include "ASgFuncs2.h"
#include "attributes.h"
#include "properties.h"
#include "error.h"

#define IN_ASGFUNCS3_C
#include "ASgFuncs3.h"


/* class Const_Expr_Checker */
/* (The next `page' (^L) is the end of the class.) */

/*
  This class traverses through an expression tree looking for
  non-constant entries. If it finds one, it sets a "bad" flag.
  The basic idea is that I want to know that (10+5)*3 is a constant
  expression, and this is a good way for me to find that out.

  Usage:

  Constant_Expr_Checker check_it (the_expr);
  check_it.check();
  int result = check_it.is_constant();

  Any expression with the CONSTANT_VALUE attribute is constant. Period.
  */


class Constant_Expr_Checker : public Iterator_Expr {
  public:
     Constant_Expr_Checker (SgExpression *the_expr);
     SgExpression *process (SgExpression *current_expr);

     void check (void);
     int is_constant(void);

  private:
     int constant_flag;
};


/* Inlines */

inline
Constant_Expr_Checker::Constant_Expr_Checker (SgExpression *the_expr)
:Iterator_Expr(the_expr)
{
     constant_flag = 1;
}

inline void
Constant_Expr_Checker::check (void)
{
     begin ();
}

inline int
Constant_Expr_Checker::is_constant(void)
{
     return constant_flag;
}

/*
  If we find an expression that is not a value, not a cast,
  not a sizeof() "call", not a set of parentheses, not an
  expression list (which is also a set of parentheses), and not
  an arithemtic op, then it is considered "non-constant"
  and the flag is set.

  I am not sure if we should approve anything, as long as it has the
  CONSTANT_VALUE attribute. The entire subtree is ignored, since it is
  presumed to have passed this test earlier.
  */

SgExpression *
Constant_Expr_Checker::process (SgExpression *current_expr)
{
     int variant = current_expr->variant();

     int do_not_check_further = 0;

     if ( 
	  ( ASgIsValExpr (current_expr)
	    || ASgIsArithmeticOpExpr (current_expr)
	    || ASgIsLogicalOpExpr (current_expr)
	    || (variant == SIZE_OP)
	    || (variant == PAREN_OP)
	    || (variant == EXPR_LIST)
	    || (variant == MOD_OP)
	    )
	  )
     {
	  /* constant_flag untouched */
     }
     else if (has_attribute (current_expr, CONSTANT_VALUE))
     {
	  do_not_check_further = 1;
	  warn ("Expression approved as constant only because it had the CONSTANT_VALUE flag.\n");
     }
     else if (variant == CAST_OP)
     {
	  /* This section could be a surprise to somebody just using this
	     code because it changes the expected definition of Constant!

	     Constant == has no gradient. An expression `(double) 3'
	     is constant. So is `(int) (x+y)' and `(double) i'.

	     */

	  Cast_Classification kind_of_cast =
	       classifyCast (current_expr);

	  if ((kind_of_cast != FLOATING_TO_SAME_FLOATING)
	      && (kind_of_cast != FLOATING_TO_OTHER_FLOATING) )
	  {
	       /* We have found a cast to a non-floating type.
		  The partial deriv across this cast is 0, so we
		  needn't look further.

		  NOTE: This says that `(int) (x+y)' is constant!
		  It also says that `(double) i' is constant.
		  */
	       do_not_check_further = 1;
	  }

     }
     else
     {
	  constant_flag = 0;
     }


     /* Prune search if we have decided it's not constant or if we know
	the rest of the tree is constant. */
     if ( (constant_flag == 0) || do_not_check_further )
	  return 0;
     else
	  return current_expr;
}


/*
  isConstantExpr:
  
  I want to know that (10+5)*3 is a constant expression. This function
  will tell me that.

  The function `isConstantExpr' checks the expression tree it gets for
  values, casts, arithmetic ops, EXPR_LIST nodes, and PAREN_OP nodes.
  If it finds something that is not one of those, it believes that the
  expression is non-constant, an returns false. If the whole tree is
  this type of node, it returns true.

  WARNING: This function considers any cast of a non-floating type
  to a floating type to be CONSTANT! This is because such an object
  will have no gradient.

  To be 100% safe using this function, interpret "constant" to me
  "is known to have no gradient".

  */

int 
isConstantExpr (SgExpression *the_expr)
{
     if ( the_expr == 0 )
	  return 0;

     Constant_Expr_Checker investigate_expr (the_expr);
     investigate_expr.check();
 
     return investigate_expr.is_constant();
}


/*
  The `Safe_Expr' class checks an expression to see if it is "safe"
  to evaluate more than once. One very nice thing this will do is
  let us just repeat lvalues that appear, rather than worrying that
  they might be the result of function calls.

  So far, the following things are safe:
       values, casts, sizeof ops, arithmetic operations, EXPR_LIST nodes,
       PAREN_OP nodes, variable references, array references,
       POINTST_OP (->), RECORD_REF (.), logical operations,
       bit shift operations, address-of operations, mod ops,
       and dereferences(!)

  We begin with the presumption that the expression is safe, and
  flag it as unsafe if we find anything bad.

 */

class Safe_Expr : public Iterator_Expr {
  public:
     Safe_Expr (SgExpression *the_expr);
     void check (void);
     int is_safe (void);

     SgExpression *process (SgExpression *current_expr);

  private:
     int safe_flag;
};


/* Inlines */

inline 
Safe_Expr::Safe_Expr (SgExpression *the_expr)
:Iterator_Expr(the_expr)
{
     /* Constructor work, and... */
     safe_flag = 1;
}

inline void
Safe_Expr::check (void)
{
     begin();
}

inline int
Safe_Expr::is_safe (void)
{
     return safe_flag;
}

/*
  Check for `unsafe' expressions: See class description

  */

SgExpression *
Safe_Expr::process (SgExpression *current_expr)
{
     int variant = current_expr->variant();
     int OK;

     /* Lump a bunch of conditions together, rather than writing
	separate `if' clauses for each. */
     
     switch (variant)
     {
       case EXPR_LIST: case PAREN_OP:
       case VAR_REF: case ARRAY_REF:
       case POINTST_OP: case RECORD_REF:
       case ADDRESS_OP:

       case DEREF_OP:

       case SIZE_OP: case CAST_OP: 

       case MOD_OP:
	  OK = 1;
	  break;
       default:
	  OK = 0;
	  break;
     }

     
     if ( ! (ASgIsValExpr (current_expr)
	     || ASgIsArithmeticOpExpr (current_expr)
	     || ASgIsLogicalOpExpr (current_expr)
	     || (OK == 1))
	  )
     {
	  safe_flag = 0;
     }

     /* Prune search if we have found bad one */
     if (safe_flag == 0)
	  return 0;
     else
	  return current_expr;
}


/*
  `isSafeExpr' checks an expression to see if it is "safe"
  to evaluate more than once. One very nice thing this will do is
  let us just repeat lvalues that appear, rather than worrying that
  they might be the result of function calls.
  
  So far, the following things are safe:
       values, casts, sizeof ops, arithmetic operations, EXPR_LIST nodes,
       PAREN_OP nodes, variable references, array references,
       POINTST_OP (->), RECORD_REF (.), logical operations,
       bit shift operations, address-of operations, and mod ops.

  We begin with the presumption that the expression is safe, and
  flag it as unsafe if we find anything bad.

 */
int
isSafeExpr (SgExpression *the_expr)
{
     Safe_Expr maybe_safe (the_expr);
     maybe_safe.check();
     return maybe_safe.is_safe();
}

     

/*
  classifyCast:

  Determine whether a cast is:
  () Cast of anything to a non-floating type.
  () Cast of a non-floating type to a floating type.
  () Cast of floating type to the same floating type. (No effect cast.)
  () Cast of one floating type to another floating type.
  [This includes cast of float to unsigned float and the like.]

  or:
  () Not a cast.
  () Cast with no children. [Appears as operand of `sizeof'.]

  */


enum Cast_Classification
classifyCast (SgExpression *the_expr)
{
     /* Trying out alan's style. */
     assert (the_expr != 0);

     if ( the_expr->variant() != CAST_OP )
	  return NOT_A_CAST;

     SgExpression *the_child = the_expr->lhs();

     if ( the_child == 0 )
	  return NO_CHILDREN;

     SgType *the_type =
	  the_expr->type()->getTrueType();
     
     if ( ! isFloatingType (the_type) )
	  return SOMETHING_TO_NON_FLOATING;

     /* From here on, we know the cast is to a floating type. */

     SgType *the_child_type =
	  the_child->type()->getTrueType();

     if ( ! isFloatingType (the_child_type) )
	  return NON_FLOATING_TO_FLOATING;

     /* Now we know both child and parent are floating types. */

     if (the_type->equivalentToType (the_child_type))
	  return FLOATING_TO_SAME_FLOATING;

     /* Not the same floating type, so it must be a different floating type. */

     return FLOATING_TO_OTHER_FLOATING;

}


