/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/****************************************************************
 *                                                              *
 *         pC++ kernel classes and SuperKernel collection       *
 *         This file should be included only after are          *
 *         element classes defined.                             *
 *                                                              *
 *         Originally written by Francois Bodin                 *
 *         Modified often by:                                   *
 *         Pete Beckman, Bernd Mohr, Shelby Yang                *
 *         Computer Science, Indiana University                 *
 *                                                              *
 ****************************************************************/

#include "GlobalStruct.h"
#include "profiler.h"
#include "pcxx_events_def.h"

// Note that "brkRuntime.h" & "brkExecDef.h" are included later in this file.
// They must be included after the class Kernel declaration.

/*******************************************************************/
/* Global Variable to ALL Run-time environments */

int pcxx_MsgBlockSize =  512;  /* size in INTS */
int pcxx_MaxMsgSize   = 16384;  /* size in INTS */
/*******************************************************************/

extern  "C" {
  void RemoteFunctionList(int* element_address,int function_number,
				    char *arg_buffer);
}

class Processors{
  public :
    int dimension;
};

class Distribution {
 public:
   int nbdim;
   int nbelement;
   int dim1size, dim2size, dim3size;
   int dist1, dist2, dist3;
   Processors P;
   Distribution();
   Distribution(int d1size, Processors *P, int dt1);
   Distribution(int d1size, int d2size, Processors *P, int dt1, int dt2);
   Distribution(int d1size, int d2size, int d3size, Processors *P, int dt1, 
	    int dt2, int dt3);
   int GiveElementProc(int i);
   int GiveElementProc(int i, int j);
   int GiveElementProc(int i, int j, int k);
};

Distribution::Distribution() {
  nbdim =1;
  nbelement = pcxx_TotalNodes();
  dim1size = pcxx_TotalNodes();
  dist1 = CYCLIC;
}

Distribution::Distribution(int d1size, Processors *, int dt1) {
  nbdim =1;
  nbelement = d1size;
  dim1size = d1size;
  dist1 = dt1;
}

Distribution::Distribution(int d1size, int d2size, Processors *, int dt1, int dt2) {
  nbdim =2;
  nbelement = d1size*d2size;
  dim1size = d1size;
  dim2size = d2size;
  dist1 = dt1;
  dist2 = dt2;
}

Distribution::Distribution(int d1size, int d2size, int d3size, Processors *, 
		   int dt1, int dt2, int dt3) {
  nbdim =2;
  nbelement = d1size*d2size*d3size;
  dim1size = d1size;
  dim2size = d2size;
  dim3size = d3size;
  dist1 = dt1;
  dist2 = dt2;
  dist3 = dt3;
  Message("Distribution(): Sorry, 3D template not implemented.",FATAL);
}
     
int Distribution::GiveElementProc(int i) {
  int proc = 0;
  int p, rem, q;
  switch (dist1) {
    case BLOCK:
/**************************************************************************
Notes from Suresh Srinivas (6/23/93):

New version of Block distribution, tries to keep the allocation
of elements to processors balanced. It tries to divide the data-structure
into contiguous blocks whose sizes are identical, except for the last
block which may be of a smaller size.  

Let `dimsize' be the size of a distributee in a certain dimension
and let `Numproc' be the number of processors in that dimension.
[Note: Integer division (i.e x/y where x and y are integers) gives the
same result as ceil((x*1.0)/(y*1.0))]

Let p   = dimsize/Numproc
    rem = dimsize - Numproc * (dimsize/Numproc)
    q = if (p <= rem) p + 1  else  p

The distribution function `delta' on the index space in that dimension is
delta(i) = { j } for all i, 0 <= i <= (dimsize - 1)
            where  j = i/q
            (note if (j >= Numproc) then j = (Numproc - 1) - j % Numproc
[Note: When j exceeds Numproc; it is better to keep j close to the computed
value by setting j to (Numproc - 1) - j % Numproc instead of trying
j = j % Numproc. The reason being that the former promotes better locality]

Example:
a) dimsize = 32, Numproc = 4.
   p = 8, rem = 0, q = 8.
   Each processor gets 8 elements.
b) dimsize = 33, Numproc = 4.
   p = 8, rem = 1, q = 8.
   Each processor gets 8 except one processor which gets 9.
c) dimsize = 99, Numproc = 50
   p = 1, rem = 49, q = 2.
   Each processor gets 2 except one processor which gets 1.
****************************************************************************/
      p = dim1size/pcxx_TotalNodes();
      rem = dim1size - pcxx_TotalNodes() * p;
      q = (p <= rem) ? p+1 : p;
      proc = i/q;
      if (proc >= pcxx_TotalNodes()) proc = (pcxx_TotalNodes() - 1) - proc % pcxx_TotalNodes();
      break;
    case CYCLIC:
      proc = (i % pcxx_TotalNodes());
      break;
    case WHOLE:
      Message("GiveElementProc()_1D: WHOLE has no meaning in 1D Distribution,",
	      WARNING);
      Message("GiveElementProc()_1D: use BLOCK distribution instead.",
	      WARNING);
      p = dim1size/pcxx_TotalNodes();
      rem = dim1size - pcxx_TotalNodes() * p;
      q = (p <= rem) ? p+1 : p;
      proc = i/q;
      if (proc >= pcxx_TotalNodes()) proc = (pcxx_TotalNodes() - 1) - proc % pcxx_TotalNodes();
      break;
    default:
      Message("GiveElementProc()_1D: Unknow distribution for 1D Distribution.", 
	      FATAL);
    }

  if (proc >= pcxx_TotalNodes()) {
    Message("GiveElementProc()_1D: Error in 1D Distribution object.",WARNING);
    proc = proc %pcxx_TotalNodes();
  }

  return proc;
}

int Distribution::GiveElementProc(int i, int j) {
  int proc = 0;
  int index1,index2, sq, p, rem, q;
  float x;

  if ((dist1 == QBLOCK) && (dist2 == QBLOCK)) {
    q = (int) QGridToProc(i,j,dim1size);
    x = dim1size*dim2size/ (float) pcxx_TotalNodes();
    proc = (int) (q/x);
    if (proc >= pcxx_TotalNodes()) {
      Message("GiveElementProc()_2D: Processor # Out of Bounds)", 
	      WARNING);
      proc = proc %pcxx_TotalNodes();
    }
    return proc;
  }
  
  sq = (int) sqrt((double)  pcxx_TotalNodes());
  switch (dist1) {
    case BLOCK:
      p = dim1size/sq;
      rem = dim1size - sq * p;
      q = (p <= rem) ? p+1 : p;
      index1 = i/q;
      if (index1 >= sq) index1 = (sq - 1) - index1 % sq;
      break;
    case CYCLIC:
      index1 = (i % sq);
      break;
    case WHOLE:
      index1 = 0;
      break;
    default:
      Message("GiveElementProc()_2D: Unknow distribution for 2D Distribution.", 
	      FATAL);
    }
  switch (dist2) {
    case BLOCK:
      p = dim2size/sq;
      rem = dim2size - sq * p;
      q = (p <= rem) ? p+1 : p;
      index2 = j/q;
      if (index2 >= sq) index2 = (sq - 1) - index2 % sq;
      break;
    case CYCLIC:
      index2 = (j % sq);
      break;
    case WHOLE:
      index2 = 0;
      break;
    default:
      Message("GiveElementProc()_2D: Unknow distribution for 2D Distribution.",
	      FATAL);
    }
  
  proc = index1 + index2*(sq);
  if (proc >= pcxx_TotalNodes()) {
    Message("GiveElementProc()_2D: Error in 2D Distribution object.", WARNING);
    proc = proc %pcxx_TotalNodes();
  }
  
  return proc;
}

int Distribution::GiveElementProc(int, int, int) {
  Message("GiveElementProc()_3D: Sorry, 3D template not implemented.",FATAL);
  return 0;
}

class Align {
  public:
    int nbelement;
    int nbdim;
    int dim1size,dim2size,dim3size;
    int ret_index1;
    int ret_index2;
    int ret_index3;
    int chosen;
    Align(int, int, char *) {
      /* fool the compiler, the real one is the next */
    };
    /* the real one */
    Align(int n1, int n2, int choix) {
      nbdim = 2;
      dim1size = n1;
      dim2size = n2;
      nbelement = n1*n2;
      chosen = choix;
    };
    Align(int n1, int choix) {
      nbdim = 1;
      dim1size = n1;
      dim2size = 0;
      nbelement = n1;
      chosen = choix;
    };
    void align_func(int var_i, int var_j); /* avoid shadow */
};

void Align::align_func(int varZZ_i, int varZZ_j)
{
  /* Set ret_index1 and ret_index2 depending on dep2C+
     generated align function (PHB) */
  GenAlignFunc(chosen,varZZ_i,varZZ_j,&ret_index1,&ret_index2);
}

class Kernel {
  public:
    int **ElementLink;  
    Distribution *Tmp;
    Align * align;
    struct INFO_HEAD GlobalInfo;
    int CollectionNumber;
    int CollectionNumber_Type; 
    int *buffer;
    Kernel(Distribution *T,  Align *A, int sizelem);
    Kernel(Kernel *col);
    ~Kernel();
    int Is_Local(int index); 
    int Is_Local(int index1, int index2); 
    int GlobalIndex(int i);   
    int GlobalIndex(int i, int j);   
    int Get_2DHighIndex(int index); 
    int Get_2DLowIndex(int index); 
    int AllocatedProc(int index);
    int AllocatedProc(int index1, int index2);
    int Install_Collection(Distribution *T, Align *A, int sizeelem);
    int *Get_Element(int index);
    int *Get_Element(int index1, int index2 );
    int *Get_ElementPart(int index,int startof, int size);
    int *Get_ElementPart(int index1, int index2,int startof, int size);
    int Kernel::Get_ElementPart_Int(int index1, int startof);
    float Kernel::Get_ElementPart_Float(int index1, int startof);
    double Kernel::Get_ElementPart_Double(int index1, int startof);
    int Kernel::Get_ElementPart_Int(int index1, int index2, int startof);
    float Kernel::Get_ElementPart_Float(int index1, int index2, int startof);
    double Kernel::Get_ElementPart_Double(int index1, int index2, int startof);
    int *Get_CopyElem(int index);
    int *Get_CopyElem(int index1, int index2 );
    void Fetch_Element(int index,int *buf,int startof, int size);
};


//  Including breakpoint stuff here...;
#ifdef BRKPNT_ON
#include "baseClasses.h"
#include "TIbasics.h"
#include "brkRuntime.h"
#endif
#include "brkExecDef.h"

int  Kernel::AllocatedProc(int index)
{
  // Assign the indexth element to a processor according to the
  // given template and alignment 
  int proc = 0;
  align->align_func(index, 0);
  if(Tmp->nbdim == 1)
    proc = Tmp->GiveElementProc(align->ret_index1);
  else
    proc = Tmp->GiveElementProc(align->ret_index1,align->ret_index2);
  return proc;
}

int Kernel::AllocatedProc(int index1, int index2)
{
  int proc = 0;
  align->align_func(index1, index2);
  proc = Tmp->GiveElementProc(align->ret_index1, align->ret_index2);
  return proc;
}

int Kernel::GlobalIndex(int index)
{
  return index; 
}

int Kernel::GlobalIndex(int index1, int index2)
{
  return (index1+index2*GlobalInfo.dim1size); 
}

int Kernel::Get_2DHighIndex(int index)
{
  return (index/GlobalInfo.dim1size);
}

int Kernel::Get_2DLowIndex(int index)
{
  return (index%GlobalInfo.dim1size);
}

int Kernel::Is_Local(int index)
{
  if(GlobalInfo.Where[GlobalIndex(index)] != pcxx_NodeNumber())
    return 0;
  else
    return 1; 
}

int Kernel::Is_Local(int index1, int index2)
{
  if(GlobalInfo.Where[GlobalIndex(index1,index2)] != pcxx_NodeNumber())
    return 0;
  else
    return 1;
}

Kernel :: Kernel(Distribution *T, Align *A, int sizelem) 
{
  if (!T || !A)
    Message("Kernel(): No alignement or template provided.",FATAL);
  Tmp = T;
  align = A;
  if(A->nbelement > T->nbelement) {
    Message("Kernel(): Elements in Align are more than elements in Distribution.",
	    WARNING);
  }
  GlobalInfo.type = -1; /* undefined */
  GlobalInfo.number_dimension = A->nbdim; /* the default */
  GlobalInfo.element_size = sizelem;
  GlobalInfo.nb_element = A->nbelement;
  GlobalInfo.dim1size  = A->dim1size;
  GlobalInfo.dim2size  = A->dim2size;
  GlobalInfo.dim3size  = A->dim3size;
  if (!Install_Collection(T,A,sizelem))
    Message("Kernel(): Error in creating collection.",FATAL);
  pcxx_Barrier();
  brk_TOGGLE_BREAKPOINT;
}

Kernel :: Kernel(Kernel *col)
{
  Distribution *T;
  Align *A;
  int sizelem;
  int i;

  if (!col)
    Message("Kernel(): Collection not provided.",FATAL);
  T = col->Tmp;
  A = col->align;
  sizelem = col->GlobalInfo.element_size;

  if (!T || !A)
    Message("Kernel(): No alignement or template provided.",FATAL);
  Tmp = T;
  align= A;

  GlobalInfo.type = -1; /* undefined */
  GlobalInfo.number_dimension = A->nbdim; /* the default */
  GlobalInfo.element_size = sizelem;
  GlobalInfo.nb_element = A->nbelement;
  GlobalInfo.dim1size  = A->dim1size;
  GlobalInfo.dim2size  = A->dim2size;
  GlobalInfo.dim3size  = A->dim3size;
  if (!Install_Collection(T,A,sizelem))
    Message("Kernel(): Error in creating collection.",FATAL);

  for (i=0; i < GlobalInfo.nb_element; i++) {
    if (GlobalInfo.Table[i] && col->GlobalInfo.Table[i]) {
      memcpy((char *) GlobalInfo.Table[i],(char *) col->GlobalInfo.Table[i], 
	     sizelem);
    }
  }
  pcxx_Barrier();
  brk_TOGGLE_BREAKPOINT;
}

/* BEGIN MACHINE DEPENDANT SECTION */

#if defined(_SMRTS_)
   /* use shared memory runtime system definitions for */
   /* BBN TC2000, Sequent Symmetry, KSR1               */
#  include "kernel-SMRTS.h"
#else
#  if defined(_AWE_)
      /* special stuff for AWESIME simulator */
#     include "kernel-AWE.h"
#  else
   /* cm5 paragon                                                  */
   /* uniprocessors are regarded as 1-node distributed systems :-) */
#  include "kernel-DMRTS.h"
#  endif
#endif

/* END MACHINE DEPENDANT FUNCTIONS */

/* START NON-MACHINE DEPENDANT FUNCTIONS */
/* non-class functions */

int GetFirstLocalElementIndex(Kernel *p)
{
  Kernel *pt;
  int i;
  pt = (Kernel *) p;

  for (i = 0; i < pt->GlobalInfo.nb_element; i++) {
    if (pt->GlobalInfo.Where[i] == pcxx_NodeNumber()) return i;
  }   
  return -1;
}

void ResetLocal(Kernel *p)
{
  p->GlobalInfo.LocalTraversalCount = 0;
  SetNextCollectionTraversalLevel((char *) p);
}

int Local (int)
{
  Kernel *pt;
  int i, node_number;
  pt = (Kernel *) GetCollectionTraversalLevel();

  if (pt->GlobalInfo.LocalTraversalCount > pt->GlobalInfo.nb_element) {
    DecCollectionTraversalLevel();
    return -1;
  }

  node_number = pcxx_NodeNumber();
  
  for (i = pt->GlobalInfo.LocalTraversalCount; i < pt->GlobalInfo.nb_element;
       i++) {
    if (pt->GlobalInfo.Where[i] == node_number) {
      pt->GlobalInfo.LocalTraversalCount = i+1;
      return i;
    }
  }   
  DecCollectionTraversalLevel();
  return -1;
}

int  FirstLocal ()
{
  return Local(0);
}


/* functions to deal with subsets for a collection with one dimension */

int FirstInSubset(Kernel *coll, int lb, int ub, int)
{
  if (!coll)
    return -1;
  if (lb == -1)
    return 0;
  if (lb <= ub)
    return lb;
  return -1;
}

int NextInSubset(Kernel *coll, int, int ub, int step, int var)
{
  int bsup;

  if (!coll)
    return -1;

  if (ub == -1)
    bsup = coll->GlobalInfo.dim1size;
  else 
    bsup = ub;

  if (var + step <= bsup)
    return var + step;
  else
    return -1;
}

CheckLocal(Kernel *coll, int index)
{
  if (!coll)
    return 0;
  
  return coll->Is_Local(index);
}

/**************************************************************************
   A kernel collection, SuperKernel, from which all new collections should 
   be derived.  It can also be used directly since basic element access and
   basic element management are provided.  Based on Francios Bodin's code
   in DistributedArray with modifications.

   Shelby Yang, Computer Science, Indiana University, 6/22/93.
***************************************************************************/
#ifdef _SMRTS_
#include <stddef.h>
#endif

Collection SuperKernel : public Kernel {
 public:
  int dim1size;
  int dim2size;
  int dimensions;
  SuperKernel(Distribution *T, Align *A);
  ~SuperKernel();
  ElementType  *operator()(int i1);           //pointer to a collection element
  ElementType  *operator()(int i1, int i2);   //pointer to a collection element
  void RemoteExecute(int index,int function_number,char *arg_buffer,int size,
		     int *tag_location);
  void RemoteExecute(int index,int function_number,char *arg_buffer,int size);
  void RemoteExecute(int index1,int index2,int function_number,
		     char *arg_buffer,int size,int *tag_location);
  void RemoteExecute(int index1,int index2,int function_number,
		     char *arg_buffer,int size);
 MethodOfElement:
  int Ident;
  int index1;
  int index2;
  SuperKernel<ElementType> *ThisCollection;
  ElementType *Self(int i);       //copy an element with an offset i 
  ElementType *Self(int i, int j);//copy an element with an offset i,j
#ifdef _SMRTS_
  void * operator new(size_t size) {
    return (void *)SMRTSmalloc(size);
  }
  void operator delete(void *p, size_t ) {
    SMRTSfree(p);
  }
#endif
};

void SuperKernel::RemoteExecute(int index,int function_number,
				char *arg_buffer,int size)
{
  /* the arguments for the remote function should be copied into the
     arg_buffer. index is the index of the element. function_number
     is the number assigned to the function that will be executed. size
     is the size of arg_buffer in bytes.
   */

  int element_index = GlobalIndex(index);

  if(Is_Local(index)) {
    RemoteFunctionList(GlobalInfo.Table[element_index],function_number,
		       arg_buffer);
    return;
  }
  AskForRemoteExecute(CollectionNumber,element_index,function_number,
		      arg_buffer,size,(long)0);
}

void SuperKernel::RemoteExecute(int index,int function_number,
				char *arg_buffer,int size,int *tag_location)
{
  int element_index = GlobalIndex(index);

  if(Is_Local(index)) {
    RemoteFunctionList(GlobalInfo.Table[element_index],function_number,
		       arg_buffer);
    (*tag_location)++;
    return;
  }
  AskForRemoteExecute(CollectionNumber,element_index,function_number,
		      arg_buffer,size,(long)tag_location);
}

void SuperKernel::RemoteExecute(int index1,int index2,int function_number,
				char *arg_buffer,int size)
{
  /* the arguments for the remote function should be copied into the
     arg_buffer. index1 and index2 are the indices of the element. 
     function_number is the number assigned to the function that will be 
     executed. size is the size of arg_buffer in bytes.
   */

  int element_index = GlobalIndex(index1,index2);

  if(Is_Local(index1,index2)) {
    RemoteFunctionList(GlobalInfo.Table[element_index],function_number,
		       arg_buffer);
    return;
  }
  AskForRemoteExecute(CollectionNumber,element_index,function_number,
		      arg_buffer,size,(long)0);
}

void SuperKernel::RemoteExecute(int index1,int index2,int function_number,
				char *arg_buffer,int size,int *tag_location)
{
  int element_index = GlobalIndex(index1,index2);

  if(Is_Local(index1,index2)) {
    RemoteFunctionList(GlobalInfo.Table[element_index],function_number,
		       arg_buffer);
    (*tag_location)++;
    return;
  }
  AskForRemoteExecute(CollectionNumber,element_index,function_number,
		      arg_buffer,size,(long)tag_location);
}

ElementType * SuperKernel :: operator()(int i1)
{
  ElementType  *pt;
  pt = (ElementType  *) Get_Element(i1);
  return pt;
}

ElementType * SuperKernel :: operator()(int i1, int i2)
{
  ElementType  *pt;
  pt = (ElementType  *) Get_Element(i1,i2);
  return pt;
}

SuperKernel::SuperKernel(Distribution *T, Align *A)
                        :Kernel(T, A, (int)sizeof(ElementType))
{
  int i, j;

  if (!T || !A)
    Message("No alignment or template for collection initialization.",FATAL);
  dim1size = A->dim1size;
  dim2size = A->dim2size;
  dimensions = A->nbdim;

  if (A->nbdim == 1) {
    for (i = 0;  i < dim1size; i++) {
      if (Is_Local(i)) {
        ElementLink[GlobalIndex(i)] = (int*) new ElementType;
      } 
    }
    pcxx_Barrier();
    for(i = 0; i < dim1size; i++) {
      GlobalInfo.Table[GlobalIndex(i)] = (int*) ElementLink[GlobalIndex(i)];
      if (Is_Local(i)) {
	(*this)(i)->index1 = i;
	(*this)(i)->index2 = 0;
	(*this)(i)->ThisCollection = (SuperKernel<ElementType> *) this;
	(*this)(i)->Ident = this->CollectionNumber*16777216+i; 
      }
    }
  }
  
  if (A->nbdim == 2) {
    for (i = 0;  i < dim1size; i++) {
      for (j = 0;  j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  ElementLink[GlobalIndex(i,j)] = (int*) new ElementType;
	}
      }
    }
    pcxx_Barrier();
    for (i = 0;  i < dim1size; i++) {
      for (j = 0;  j < dim2size; j++) {
	GlobalInfo.Table[GlobalIndex(i,j)] = (int*) ElementLink[GlobalIndex(i,j)];
        if (Is_Local(i,j)) {
          (*this)(i,j)->index1 = i;
          (*this)(i,j)->index2 = j;
          (*this)(i,j)->ThisCollection = (SuperKernel<ElementType> *) this;
          (*this)(i,j)->Ident = this->CollectionNumber*16777216+i+j*this->dim1size;
        } 
      }
    }
  }

  if (A->nbdim >= 3)
    Message("3D and higher-dimensional collection not implemented.",FATAL);

#ifdef _SMRTS_
  if(pcxx_NodeNumber() == 0) SMRTSDealloc(ElementLink);
#else
  delete [] ElementLink;
#endif
}

SuperKernel::~SuperKernel() 
{
  int i, j;
  if (dimensions == 1) {
    for (i = 0;  i < dim1size; i++) {
      if (AllocatedProc(i) == pcxx_NodeNumber()) {
        delete (ElementType*) GlobalInfo.Table[GlobalIndex(i)];
      } 
    }
  }

  if (dimensions == 2) {
    for (i = 0;  i < dim1size; i++) {
      for (j = 0;  j < dim2size; j++) {
        if (AllocatedProc(i, j) == pcxx_NodeNumber()) {
          delete (ElementType*) GlobalInfo.Table[GlobalIndex(i,j)];
        } 
      }
    }
  }

  if (dimensions >= 3)
    Message("3D and higher-dimensional collection not implemented.",FATAL);
}


ElementType * SuperKernel :: Self(int i)
{
  return (ElementType  *) ThisCollection->Get_Element(index1 + i);
}

ElementType * SuperKernel :: Self(int i, int j)
{
  return (ElementType  *) ThisCollection->Get_Element(index1 + i, index2 + j);
};

